import requests
from future.moves.urllib.parse import urljoin

from omniduct.duct import Duct


class RestClientBase(Duct):
    """
    This is a simple wrapper around the requests library to simplify the use
    of RESTful clients with omniduct. This allows all the automatic features
    around port forwarding from remote hosts to be inherited. This client can
    be used directly, or decorated by subclasses which can add methods
    specific to any REST service; and internally use `request` and `request_json`
    to access various endpoints.
    """

    DUCT_TYPE = Duct.Type.RESTFUL

    def __init__(self, server_protocol='http', assume_json=False, endpoint_prefix='', **kwargs):
        """
        This is a shim __init__ function that passes all arguments onto
        `self._init`, which is implemented by subclasses. This allows subclasses
        to instantiate themselves with arbitrary parameters.
        """
        Duct.__init_with_kwargs__(self, kwargs, port=80)

        self.server_protocol = server_protocol
        self.assume_json = assume_json
        self.endpoint_prefix = endpoint_prefix

        self._init(**kwargs)

    def _init(self):
        pass

    def __call__(self, endpoint, method='get', **kwargs):
        if self.assume_json:
            return self.request_json(endpoint, method=method, **kwargs)
        return self.request(endpoint, method=method, **kwargs)

    @property
    def base_url(self):
        url = urljoin('{}://{}:{}'.format(self.server_protocol, self.host, self.port or 80), self.endpoint_prefix)
        if not url.endswith('/'):
            url += '/'
        return url

    def request(self, endpoint, method='get', **kwargs):
        self.connect()
        url = urljoin(self.base_url, endpoint)
        return requests.request(method, url, **kwargs)

    def request_json(self, endpoint, method='get', **kwargs):
        request = self.request(endpoint, method=method, **kwargs)
        if not request.status_code == 200:
            raise RuntimeError("Server responded with HTTP response code {}, with content: {}.".format(request.status_code, request.content.decode()))
        return request.json()

    def _connect(self):
        pass

    def _is_connected(self):
        return True

    def _disconnect(self):
        pass


class RestClient(RestClientBase):
    PROTOCOLS = ['rest']



import unittest
import mock
from pyfakefs.fake_filesystem_unittest import Patcher

from omniduct.caches.local import LocalCache


ID_DUCT = 'test_id_duct'
ID_STRING = 'test_id_string'
ID_STRING_ANOTHER = 'test_id_string_another'
ID_STRING_NONEXISTANT = 'test_id_string_nonexistant'
TEST_DIR = 'test_dir'


class TestLocalCache(unittest.TestCase):

    def setUp(self):
        self.fs_patcher = Patcher()
        self.fs_patcher.setUp()
        self.cache = LocalCache(TEST_DIR)

    def test_clear(self):
        self.cache.clear(ID_DUCT, ID_STRING_NONEXISTANT)
        self.cache.set(ID_DUCT, ID_STRING, 'foo')
        self.cache.clear(ID_DUCT, ID_STRING)
        self.assertFalse(
            self.cache.has_key(ID_DUCT, ID_STRING),
            'expected not to find key if key is cleared'
        )
        self.assertIsNone(
            self.cache.get(ID_DUCT, ID_STRING),
            'expected to get None if key has been cleared'
        )

    def test_clear_all(self):
        self.cache.set(ID_DUCT, ID_STRING, 'foo')
        self.cache.set(ID_DUCT, ID_STRING_ANOTHER, 'bar')
        self.cache.clear_all(ID_DUCT)

        self.assertFalse(
            self.cache.has_key(ID_DUCT, ID_STRING),
            'expected not to find key after clear_all'
        )
        self.assertFalse(
            self.cache.has_key(ID_DUCT, ID_STRING_ANOTHER),
            'expected not to find key after clear_all'
        )

    def test_get(self):
        self.assertIsNone(
            self.cache.get(ID_DUCT, ID_STRING_NONEXISTANT),
            'expected not find non-existant key'
        )
        self.cache.set(ID_DUCT, ID_STRING, 'foo')
        self.assertEqual(
            self.cache.get(ID_DUCT, ID_STRING), 'foo',
            'expected object retrieved from cache to be equal to its pre-cached value'
        )

    def test_has_key(self):
        self.assertFalse(
            self.cache.has_key(ID_DUCT, ID_STRING_NONEXISTANT),
            'expected not to find non-existant key'
        )
        self.cache.set(ID_DUCT, ID_STRING, 'foo')
        self.assertTrue(
            self.cache.has_key(ID_DUCT, ID_STRING),
            'expected to find key after setting it'
        )

    def test_keys(self):
        pass  # TODO test keys when implemented

    def test_set(self):
        serializer_mock = mock.Mock()
        self.cache.set(ID_DUCT, ID_STRING, 'foo', serializer=serializer_mock.serialize)
        serializer_mock.serialize.assert_called_once()

    def tearDown(self):
        self.fs_patcher.tearDown()



import unittest
import mock

from omniduct.restful.base import RestClient


class TestRestClient(unittest.TestCase):

    @mock.patch.object(RestClient, 'connect')
    @mock.patch('omniduct.restful.base.requests.request')
    def test_default_request(self, mock_request, mock_connect):
        client = RestClient(server_protocol='http', host='localhost', port=80)
        client.request('/')
        mock_connect.assert_called_with()
        mock_request.assert_called_with("get", "http://localhost:80/")



import datetime
import getpass
import os
import posixpath
import re
import tempfile

import pandas as pd

from omniduct.filesystems.base import FileSystemFileDesc
from omniduct.remotes.base import RemoteClient
from omniduct.utils.debug import logger
from omniduct.utils.processes import run_in_subprocess
from omniduct.errors import DuctAuthenticationError

SSH_ASKPASS = '{omniduct_dir}/utils/ssh_askpass'.format(omniduct_dir=os.path.dirname(__file__))
SESSION_SSH_USERNAME = None
SESSION_REMOTE_HOST = None
SESSION_SSH_ASKPASS = False


class SSHClient(RemoteClient):
    """SSHClient manages a persistent connection to remote hosts allowing
    remote execution, port forwarding, file transfer, and cleanup

    References
    ----------
    https://puppetlabs.com/blog/speed-up-ssh-by-reusing-connections

    """

    PROTOCOLS = ['ssh']

    def _init(self):
        pass

    # Duct connection implementation

    def _connect(self):
        """
        Create persistent connection to remote host.

        The workflow to handle passwords and host keys is inspired by the pxssh module of pexpect
        (https://github.com/pexpect/pexpect). We have adjusted this workflow to our purposes.

        Returns
        -------
        proc : Popen subprocess
            Subprocess used to connect.
        """
        import pexpect

        # Create socket directory if it doesn't exist.
        socket_dir = os.path.dirname(self._socket_path)
        if not os.path.exists(socket_dir):
            os.makedirs(socket_dir)
        # Create persistent master connection and exit.
        cmd = ("ssh {login} -MT "
               "-S {socket} "
               "-o ControlPersist=yes "
               "-o StrictHostKeyChecking=no "
               "-o NoHostAuthenticationForLocalhost=yes "
               "-o ServerAliveInterval=60 "
               "-o ServerAliveCountMax=2 "
               "'exit'".format(login=self._login_info, socket=self._socket_path))

        expected = [
            "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!",    # 0
            "(?i)are you sure you want to continue connecting",    # 1
            "(?i)(?:(?:password)|(?:passphrase for key)):",        # 2
            "(?i)permission denied",                               # 3
            "(?i)terminal type",                                   # 4
            pexpect.TIMEOUT,                                       # 5
            "(?i)connection closed by remote host",                # 6
            "(?i)could not resolve hostname",                      # 7
            pexpect.EOF                                            # 8
        ]

        try:
            expect = pexpect.spawn(cmd)
            i = expect.expect(expected, timeout=10)

            # First phase
            if i == 0:  # If host identification changed, arrest any further attempts to connect
                raise RuntimeError('Host identification for {} has changed! If you are sure that the host identification SHOULD have changed, and are not experiencing a man-in-the-middle attack, please remove the line corresponding to this server in ~/.ssh/known_hosts; or call the `update_host_keys` method of this object.'.format(self._host))
            if i == 1:  # Request to authorize host certificate (i.e. host not in the 'known_hosts' file)
                expect.sendline("yes")
                i = self.expect(expected)
            if i == 2:  # Request for password/passphrase
                expect.sendline(self.password or getpass.getpass('Password: '))
                i = self.expect(expected)
            if i == 4:  # Request for terminal type
                expect.sendline('ascii')
                i = self.expect(expected)

            # Second phase
            if i == 1:  # Another request to authorize host certificate (i.e. host not in the 'known_hosts' file)
                raise RuntimeError('Received a second request to authorize host key. This should not have happened!')
            elif i in (2, 3):  # Second request for password/passphrase or rejection of credentials. For now, give up.
                raise DuctAuthenticationError('Invalid username and/or password, or private key is not unlocked.')
            elif i == 4:  # Another request for terminal type.
                raise RuntimeError('Received a second request for terminal type. This should not have happened!')
            elif i == 5:  # Timeout
                # In our instance, this means that we have not handled some or another aspect of the login procedure.
                # Since we are expecting an EOF when we have successfully logged in, hanging means that the SSH login
                # procedure is waiting for more information. Since we have no more to give, this means our login
                # was unsuccessful.
                raise RuntimeError('SSH client seems to be awaiting more information, but we have no more to give. The '
                                   'messages received so far are:\n{}'.format(expect.before))
            elif i == 6:  # Connection closed by remote host
                raise RuntimeError("Remote closed SSH connection")
            elif i == 7:
                raise RuntimeError("Cannot connect to {} on your current network connection".format(self.host))
        finally:
            expect.close()

        # We should be logged in at this point, but let us make doubly sure
        assert self.is_connected(), 'Unexpected failure to establish a connection with the remote host with command: \n ' \
                                    '{}\n\n Please report this!'.format(cmd)

    def _is_connected(self):
        """
        Return whether SSHClient is connected by checking the control socket.
        """
        cmd = "ssh {login} -T -S {socket} -O check".format(login=self._login_info,
                                                           socket=self._socket_path)
        proc = run_in_subprocess(cmd)

        if proc.returncode != 0:
            if os.path.exists(self._socket_path):
                os.remove(self._socket_path)
            return False
        return True

    def _disconnect(self):
        """
        Exit persistent connection to remote host.
        """
        # Send exit request to control socket.
        cmd = "ssh {login} -T -S {socket} -O exit".format(login=self._login_info,
                                                          socket=self._socket_path)
        run_in_subprocess(cmd)

    # RemoteClient implementation

    def _execute(self, cmd, **kwargs):
        """
        Execute a command on a remote host.

        Parameters
        ----------
        cmd : string
            Command to be executed on remote host.
        kwargs : keywords
            Options to pass to subprocess.Popen.

        Returns
        -------
        proc : Popen subprocess
            Subprocess used to run remote command.
        """
        template = 'ssh {login} -T -o ControlPath={socket} << EOF\n{cmd}\nEOF'
        config = dict(self._subprocess_config)
        config.update(kwargs)
        return run_in_subprocess(template.format(login=self._login_info,
                                                 socket=self._socket_path,
                                                 cmd=cmd),
                                 check_output=True,
                                 **config)

    def _port_forward_start(self, local_port, remote_host, remote_port):
        self.connect()
        logger.info('Establishing port forward...')
        cmd_template = 'ssh {login} -T -O forward -S {socket} -L localhost:{local_port}:{remote_host}:{remote_port}'
        cmd = cmd_template.format(login=self._login_info,
                                  socket=self._socket_path,
                                  local_port=local_port,
                                  remote_host=remote_host,
                                  remote_port=remote_port)
        proc = run_in_subprocess(cmd)
        if proc.returncode != 0:
            raise Exception('Unable to port forward with command: {}'.format(cmd))
        logger.info(proc.stderr or 'Success')
        return proc

    def _port_forward_stop(self, local_port, remote_host, remote_port, connection):
        logger.info('Cancelling port forward...')
        cmd_template = 'ssh {login} -T -O cancel -S {socket} -L localhost:{local_port}:{remote_host}:{remote_port}'
        cmd = cmd_template.format(login=self._login_info,
                                  socket=self._socket_path,
                                  local_port=local_port,
                                  remote_host=remote_host,
                                  remote_port=remote_port)
        proc = run_in_subprocess(cmd)
        logger.info('Success' if proc.returncode == 0 else 'Failure')

    def _is_port_bound(self, host, port):
        return self.execute('which nc; if [ $? -eq 0 ]; then nc -z -w2 {} {}; fi'.format(host, port)).returncode == 0

    # FileSystem methods

    # Path properties and helpers

    def _path_home(self):
        return '~'

    def _path_separator(self):
        return '/'

    # File node properties

    def _exists(self, path):
        return self.execute('if [ ! -e {} ]; then exit 1; fi'.format(path)).returncode == 0

    def _isdir(self, path):
        return self.execute('if [ ! -d {} ]; then exit 1; fi'.format(path)).returncode == 0

    def _isfile(self, path):
        return self.execute('if [ ! -f {} ]; then exit 1; fi'.format(path)).returncode == 0

    # Directory handling and enumeration

    def _dir(self, path):
        dir = pd.DataFrame(sorted([re.split('\s+', f) for f in self.execute('ls -Al {}'.format(path)).stdout.decode().strip().split('\n')[1:]]),
                           columns=['file_mode', 'link_count', 'owner', 'group', 'bytes', 'month', 'day', 'time', 'path'])

        def convert_to_datetime(x):
            months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
            year = datetime.datetime.now().year if ':' in x.time else x.time
            time = x.time if ':' in x.time else None
            return datetime.datetime(
                year=int(year),
                month=months.index(x.month) + 1,
                day=int(x.day),
                hour=int(time.split(':')[0]) if time is not None else 0,
                minute=int(time.split(':')[1]) if time is not None else 0
            )

        if len(dir) == 0:  # Directory is empty
            return

        dir = dir.assign(
            last_modified=lambda x: x.apply(convert_to_datetime, axis=1),
            type=lambda x: x.apply(lambda x: 'directory' if x.file_mode.startswith('d') else 'file', axis=1)
        ).drop(
            ['month', 'day', 'time'],
            axis=1
        ).sort_values(
            ['type', 'path']
        ).reset_index(drop=True)

        for i, row in dir.iterrows():
            yield FileSystemFileDesc(
                fs=self,
                path=posixpath.join(path, row.path),
                name=row.path,
                type='directory' if row.file_mode.startswith('d') else 'file',
                bytes=row.bytes,
                owner=row.owner,
                group=row.group,
                last_modified=row.last_modified,
            )

    def _mkdir(self, path, recursive):
        assert self.execute('mkdir ' + ('-p ' if recursive else '') + '"{}"'.format(path)).returncode == 0, "Failed to create directory at: `{}`".format(path)

    # File handling

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        read = self.execute('cat {}'.format(path)).stdout
        if not binary:
            read = read.decode()
        return read

    def _file_append_(self, path, s, binary):
        raise NotImplementedError

    def _file_write_(self, path, s, binary):
        if binary:
            fd, tmp_path = tempfile.mkstemp()
        else:
            fd, tmp_path = tempfile.mkstemp(text=True)
        os.close(fd)

        with open(tmp_path, 'w' + ('b' if binary else '')) as f:
            f.write(s)

        return self._copy_from_local(tmp_path, path, overwrite=True)

    # File transfer

    def download(self, source, dest=None, overwrite=False, fs=None):
        """
        This method overloads the default download method, and handles
        remote-to-local downloads specially.
        """
        from ..filesystems.local import LocalFsClient

        if fs is None or isinstance(fs, LocalFsClient):
            self.connect()
            logger.info('Copying file to local...')
            dest = dest or posixpath.basename(source)
            cmd = (
                "scp -r -o ControlPath={socket} {login}:'{remote_file}' '{local_file}'".format(
                    socket=self._socket_path,
                    login=self._login_info,
                    remote_file=dest.replace('"', r'\"'),
                    local_file=source.replace('"', r'\"'),  # quote escaped for bash
                )
            )
            proc = run_in_subprocess(cmd, check_output=True)
            logger.info(proc.stderr or 'Success')
        else:
            return super(RemoteClient, self).download(source, dest, overwrite, fs)

    def upload(self, source, dest=None, overwrite=False, fs=None):
        """
        This method overloads the default download method, and handles
        local-to-remote uploads specially.
        """
        from ..filesystems.local import LocalFsClient

        if fs is None or isinstance(fs, LocalFsClient):
            self.connect()
            logger.info('Copying file from local...')
            dest = dest or posixpath.basename(source)
            cmd = (
                "scp -r -o ControlPath={socket} '{local_file}' {login}:'{remote_file}'".format(
                    socket=self._socket_path,
                    local_file=source.replace('"', r'\"'),  # quote escaped for bash
                    login=self._login_info,
                    remote_file=dest.replace('"', r'\"'),
                )
            )
            proc = run_in_subprocess(cmd, check_output=True)
            logger.info(proc.stderr or 'Success')
        else:
            return super(RemoteClient, self).upload(source, dest, overwrite, fs)

    # Helper methods

    @property
    def _login_info(self):
        return '@'.join([self.username, self.host])

    @property
    def _socket_path(self):
        # On Linux the maximum socket path length is 108 characters, and on Mac OS X it is 104 characters, including
        # the final sentinel character (or so it seems). SSH appends a '.' character, followed by random sequence of 16
        # characters. We therefore need the rest of the path to be less than 86 characters.
        return os.path.expanduser('~/.ssh/omniduct/{}'.format(self._login_info))[:86]

    @property
    def _subprocess_config(self):
        return {}

    def update_host_keys(self):
        assert not self.remote, "Updating host key only works for local connections."
        cmd = "ssh-keygen -R {host} && ssh-keyscan {host} >> ~/.ssh/known_hosts".format(host=self.host)
        return run_in_subprocess(cmd, False).returncode == 0



import datetime
import errno
import os

from .base import FileSystemClient, FileSystemFileDesc


class LocalFsClient(FileSystemClient):

    PROTOCOLS = ['localfs']
    DEFAULT_PORT = 22

    def _init(self, cwd_as_home=True):
        self.__cwd_as_home = cwd_as_home

    def _connect(self):
        pass

    def _is_connected(self):
        return True

    def _disconnect(self):
        pass

    # File enumeration
    def _path_home(self):
        return os.getcwd() if self.__cwd_as_home else os.path.expanduser('~')

    def _path_separator(self):
        return os.path.sep

    def _exists(self, path):
        return os.path.exists(path)

    def _isdir(self, path):
        return os.path.isdir(path)

    def _isfile(self, path):
        return os.path.isfile(path)

    def _dir(self, path):
        if not os.path.isdir(path):
            raise RuntimeError("No such folder.")
        for f in os.listdir(path):
            f_path = os.path.join(path, f)

            attrs = {}

            if os.name == 'posix':
                import grp
                import pwd

                stat = os.stat(f_path)

                attrs.update({
                    'owner': pwd.getpwuid(stat.st_uid).pw_name,
                    'group': grp.getgrgid(stat.st_gid).gr_name,
                    'permissions': oct(stat.st_mode),
                    'created': str(datetime.datetime.fromtimestamp(stat.st_ctime)),
                    'last_modified': str(datetime.datetime.fromtimestamp(stat.st_mtime)),
                    'last_accessed': str(datetime.datetime.fromtimestamp(stat.st_atime)),
                })

            yield FileSystemFileDesc(
                fs=self,
                path=f_path,
                name=f,
                type='directory' if os.path.isdir(f_path) else 'file',
                bytes=os.path.getsize(f_path),
                **attrs
            )

    def _walk(self, path):
        return os.walk(path)

    def _mkdir(self, path, recursive):
        try:
            os.makedirs(path) if recursive else os.makedir(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    # File opening

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        with open(path, 'r{}'.format('b' if binary else '')) as f:
            return f.read()

    def _file_append_(self, path, s, binary):
        raise NotImplementedError

    def _file_write_(self, path, s, binary):
        with open(path, 'w' + ('b' if binary else '')) as f:
            return f.write(s)



from __future__ import absolute_import

from omniduct.utils.debug import logger

from .base import DatabaseClient


class DruidClient(DatabaseClient):

    PROTOCOLS = ['druid']
    DEFAULT_PORT = 80

    def _init(self):
        self.__druid = None

    # Connection
    def _connect(self):
        from pydruid.db import connect
        logger.info('Connecting to Druid database ...')
        self.__druid = connect(self.host, self.port, path='/druid/v2/sql/', scheme='http')

    def _is_connected(self):
        return self.__druid is not None

    def _disconnect(self):
        logger.info('Disconnecting from Druid database ...')
        try:
            self.__druid.close()
        except Exception:
            pass
        self.__druid = None

    # Querying
    def _execute(self, statement, cursor=None, async=False):
        cursor = cursor or self.__druid.cursor()
        cursor.execute(statement)
        return cursor

    def _table_list(self, schema=None, like=None, **kwargs):
        cmd = "SELECT * FROM INFORMATION_SCHEMA.TABLES"
        return self.query(cmd, **kwargs)

    def _table_exists(self, table, schema=None):
        return (self.table_list(renew=True, schema=schema)['TABLE_NAME'] == table).any()

    def _table_desc(self, table, **kwargs):
        query = ("""
            SELECT
                TABLE_SCHEMA
                , TABLE_NAME
                , COLUMN_NAME
                , ORDINAL_POSITION
                , COLUMN_DEFAULT
                , IS_NULLABLE
                , DATA_TYPE
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE TABLE_NAME = '{}'""").format(table)
        return self.query(query, **kwargs)

    def _table_head(self, table, n=10, **kwargs):
        return self.query("SELECT * FROM {} LIMIT {}".format(table, n), **kwargs)

    def _table_props(self, table, **kwargs):
        raise NotImplementedError



import os

import IPython

from omniduct.registry import DuctRegistry

from .utils.config import config

# The default console width is too wide for most notebooks, leading to ugly logging message / progress bars. We set this
# to a more reasonable value for Jupyter Notebooks.
ip = IPython.get_ipython()
if ip and ip.__class__.__name__ == "ZMQInteractiveShell":
    os.environ['COLUMNS'] = "80"

__all__ = ['config', 'registry', '__author__', '__author_email__', '__version__']

OMNIDUCT_CONFIG = os.environ.get('OMNIDUCT_CONFIG', None) or os.path.expanduser('~/.omniduct/config')

config.register("ducts", "The ducts to register with the system.")

config._config_path = OMNIDUCT_CONFIG

registry = DuctRegistry(getattr(config, 'ducts', {}))
registry.populate_namespace(globals())



import re

from .base import FileSystemClient, FileSystemFileDesc
from ..utils.debug import logger


class S3Client(FileSystemClient):

    PROTOCOLS = ['s3']
    DEFAULT_PORT = 80

    def _init(self, bucket=None, aws_profile='default'):
        # Note: aws_profile, if specified, should be the name of a profile as
        # specified in ~/.aws/credentials. Authentication is handled by the
        # `opinel` library, which is also aware of environment variables.
        # Set up your command line aws client, and if it works, this should too.
        assert bucket is not None, 'S3 Bucket must be specified using the `bucket` kwarg.'
        self.bucket = bucket
        self.aws_profile = aws_profile

        self._setup_session()  # Ensure self.host is updated with correct AWS region

    def _setup_session(self):
        import boto3
        s = boto3.Session(profile_name=self.aws_profile)
        self._session = s
        self._client = s.client('s3')
        self._resource = s.resource('s3')
        self.host = 'autoscaling.{}.amazonaws.com'.format(self._client.meta.region_name)

    def _connect(self):
        from opinel.utils.credentials import read_creds

        # Refresh access token, and attach credentials to current object for debugging
        self._credentials = read_creds(self.aws_profile)

        # Update AWS session, client and resource objects
        self._setup_session()

    def _is_connected(self):
        # Check if still able to perform requests against AWS
        import botocore
        try:
            self._client.list_buckets()
        except botocore.exceptions.ClientError as e:
            if len(e.args) > 0:
                if 'ExpiredToken' in e.args[0] or 'InvalidToken' in e.args[0]:
                    return False
                elif 'AccessDenied' in e.args[0]:
                    return True

    def _disconnect(self):
        pass

    # Path properties and helpers

    def _path_home(self):
        return '/'

    def _path_separator(self):
        return '/'

    def _path(self, path):
        path = super(S3Client, self)._path(path)
        if path.startswith('/'):
            path = path[1:]
        return path

    # File node properties

    def _exists(self, path):
        return self.isfile(path) or self.isdir(path)

    def _isdir(self, path):
        response = next(iter(self.__dir_paginator(path)))
        if 'CommonPrefixes' in response or 'Contents' in response:
            return True
        return False

    def _isfile(self, path):
        try:
            self._client.get_object(Bucket=self.bucket, Key=path or '')
            return True
        except:
            return False

    # Directory handling and enumeration

    def __dir_paginator(self, path):
        if path.endswith('/'):
            path = path[:-1]
        paginator = self._client.get_paginator('list_objects')
        iterator = paginator.paginate(
            Bucket=self.bucket,
            Prefix=path + ('/' if path else ''),
            Delimiter='/',
            PaginationConfig={'PageSize': 500}
        )
        return iterator

    def _dir(self, path):
        iterator = self.__dir_paginator(path)

        for response_data in iterator:
            for prefix in response_data.get('CommonPrefixes', []):
                yield FileSystemFileDesc(
                    fs=self,
                    path=prefix['Prefix'][:-1],
                    name=prefix['Prefix'][:-1].split('/')[-1],  # Remove trailing slash
                    type='directory',
                )
            for prefix in response_data.get('Contents', []):
                yield FileSystemFileDesc(
                    fs=self,
                    path=prefix['Key'],
                    name=prefix['Key'].split('/')[-1],
                    type='file',
                    bytes=prefix['Size'],
                    owner=prefix['Owner']['DisplayName'],
                    last_modified=prefix['LastModified']
                )

    # TODO: Interestingly, directly using Amazon S3 methods seems slower than generic approach. Hypothesis: keys is not async.
    # def _find(self, path_prefix, **attrs):
    #     if len(set(attrs).difference(('name',))) > 0 or hasattr(attrs.get('name'), '__call__'):
    #         logger.warning('Falling back to recursive search, rather than using S3, since find requires filters on more than just name.')
    #         for result in super(S3Client, self)._find(path_prefix, **attrs):
    #             yield result
    #
    #     pattern = re.compile(attrs.get('name') or '.*')
    #
    #     b = self._resource.Bucket(self.bucket)
    #     keys = b.objects.filter(Prefix=path_prefix)
    #     for k in keys:
    #         if pattern is None or pattern.match(k.key[len(path_prefix):]):
    #             yield k.key

    def _mkdir(self, path, recursive):
        raise NotImplementedError("Support for S3 write operations has yet to be implemented.")

    # File handling

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        assert self.isfile(path), "File `{}` does not exist.".format(path)
        obj = self._resource.Object(self.bucket, path)
        body = obj.get()['Body'].read()

        if not binary:
            body = body.decode()

        return body

    def _file_append_(self, path, s, binary):
        raise NotImplementedError("Support for S3 write operations has yet to be implemented.")

    def _file_write_(self, path, s, binary):
        raise NotImplementedError("Support for S3 write operations has yet to be implemented.")



import posixpath

from .base import FileSystemClient, FileSystemFileDesc


class WebHdfsClient(FileSystemClient):

    PROTOCOLS = ['webhdfs']
    DEFAULT_PORT = 50070

    def _init(self, namenodes=None, global_writes=False, **kwargs):
        self._namenodes = namenodes
        self.global_writes = global_writes

        self.__webhdfs = None
        self.__webhdfs_kwargs = kwargs
        self.__home_directory = None
        self.prepared_fields += ('_namenodes',)

    def _connect(self):
        from .webhdfs_helpers import OmniductPyWebHdfsClient
        self.__webhdfs = OmniductPyWebHdfsClient(host=self._host, port=self._port, remote=self.remote, namenodes=self._namenodes, user_name=self.username, **self.__webhdfs_kwargs)
        self.__home_directory = self.__webhdfs.get_home_directory()

    def _is_connected(self):
        try:
            if self.remote and not self.remote.is_connected():
                return False
            return self.__webhdfs is not None
        except:
            return False

    def _disconnect(self):
        self.__webhdfs = None
        self.__home_directory = None

    # Path properties and helpers

    def _path_home(self):
        if not self.__home_directory:
            self.connect()
        return self.__home_directory

    def _path_separator(self):
        return '/'

    def __in_home_directory(self, path):
        return posixpath.abspath(path).startswith(self.__home_directory)

    # File node properties

    def _exists(self, path):
        from pywebhdfs.errors import FileNotFound
        try:
            self.__webhdfs.get_file_dir_status(path)
            return True
        except FileNotFound:
            return False

    def _isdir(self, path):
        from pywebhdfs.errors import FileNotFound
        try:
            stats = self.__webhdfs.get_file_dir_status(path)
            return stats['FileStatus']['type'] == 'DIRECTORY'
        except FileNotFound:
            return False

    def _isfile(self, path):
        from pywebhdfs.errors import FileNotFound
        try:
            stats = self.__webhdfs.get_file_dir_status(path)
            return stats['FileStatus']['type'] == 'FILE'
        except FileNotFound:
            return False

    # Directory handling and enumeration

    def _dir(self, path):
        files = self.__webhdfs.list_dir(path)
        for f in files['FileStatuses']['FileStatus']:
            yield FileSystemFileDesc(
                fs=self,
                path=posixpath.join(path, f['pathSuffix']),
                name=f['pathSuffix'],
                type=f['type'].lower(),
                bytes=f['length'],
                owner=f['owner'],
                group=f['group'],
                last_modified=f['modificationTime'],
                last_accessed=f['accessTime'],
                permissions=f['permission'],
                replication=f['replication']
            )

    def _mkdir(self, path, recursive):
        raise NotImplementedError

    # File handling

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        read = self.__webhdfs.read_file(path, offset=offset, length='null' if size < 0 else size)
        if not binary:
            read = read.decode()
        return read

    def _file_append_(self, path, s, binary):
        return self.__webhdfs.append_file(path, s)

    def _file_write_(self, path, s, binary):
        if not self.global_writes and not self.__in_home_directory(path):
            raise RuntimeError("Attempting to write outside of home directory without setting '{name}.global_writes' to True.".format(name=self.name))
        return self.__webhdfs.create_file(path, s, overwrite=True)
